@@ -12,6 +12,9 @@
 
 #include "config.h"
 
+#include <assert.h>
+
+#include "keymap.h"
 #include "xkbcomp-priv.h"
 #include "text.h"
 
@@ -26,13 +29,11 @@ struct buf {
 static bool
 do_realloc(struct buf *buf, size_t at_least)
 {
-    char *new;
-
     buf->alloc += BUF_CHUNK_SIZE;
     if (at_least >= BUF_CHUNK_SIZE)
         buf->alloc += at_least;
 
-    new = realloc(buf->buf, buf->alloc);
+    char *const new = realloc(buf->buf, buf->alloc);
     if (!new)
         return false;
 
@@ -90,6 +91,107 @@ err:
         return false; \
 } while (0)
 
+static bool
+check_copy_to_buf(struct buf *buf, const char* source, size_t len)
+{
+    if (len == 0)
+        return true;
+
+    const size_t available = buf->alloc - buf->size;
+    if (len >= available) {
+        /* len + 1 (terminating NULL) */
+        if (!do_realloc(buf, len + 1)) {
+            free(buf->buf);
+            buf->buf = NULL;
+            return false;
+        }
+    }
+
+    memcpy(buf->buf + buf->size, source, len);
+    buf->size += len;
+    /* Append NULL byte */
+    buf->buf[buf->size] = '\0';
+    return true;
+}
+
+#define copy_to_buf_len(buf, source, len) do { \
+    if (!check_copy_to_buf(buf, source, len)) \
+        return false; \
+} while (0)
+
+#define copy_to_buf(buf, source) \
+    copy_to_buf_len(buf, source, sizeof(source) - 1)
+
+static bool
+check_write_string_literal(struct buf *buf, const char* string)
+{
+    copy_to_buf(buf, "\"");
+
+    /* Write chunks, separated by characters requiring escape sequence */
+    size_t pending_start = 0;
+    size_t current = 0;
+    char escape_buffer[] = {'\\', '0', '0', '0', '0'};
+    for (; string[current] != '\0'; current++) {
+        uint8_t escape_len;
+        switch (string[current]) {
+        case '\n':
+            /* `\n` would break strings */
+            escape_buffer[1] = 'n';
+            escape_len = 2;
+            break;
+        case '\\':
+            /* `\` would create escape sequences */
+            escape_buffer[1] = '\\';
+            escape_len = 2;
+            break;
+        default:
+            /*
+             * Handle `"` (would break strings) and ASCII control characters
+             * with an octal escape sequence. Xorg xkbcomp does not parse the
+             * escape sequence `\"` nor xkbcommon < 1.9.0, so in order to be
+             * backward compatible we must use the octal escape sequence in
+             * xkbcomp style `\0nnn` with *4* digits:
+             *
+             * 1. To be compatible with Xorg xkbcomp.
+             * 2. To avoid issues with the next char: e.g. "\00427" should not
+             *    be emitted as "\427" nor "\0427".
+             *
+             * Note that xkbcommon < 1.9.0 will still not parse this correctly,
+             * although it will not raise an error: e.g. the escape for `"`,
+             * `\0042`, would be parsed as `\004` + `2`.
+             */
+            if (unlikely((unsigned char)string[current] < 0x20U ||
+                         string[current] == '"')) {
+                escape_buffer[1] = '0';
+                escape_buffer[2] = '0';
+                escape_buffer[3] = (char)('0' + (string[current] >> 3));
+                escape_buffer[4] = (char)('0' + (string[current] & 0x7));
+                escape_len = 5;
+            } else {
+                /* Expand the chunk */
+                continue;
+            }
+        }
+        /* Write pending chunk */
+        assert(current >= pending_start);
+        copy_to_buf_len(buf, string + pending_start, current - pending_start);
+        /* Write escape sequence */
+        copy_to_buf_len(buf, escape_buffer, escape_len);
+        pending_start = current + 1;
+    }
+    /* Write remaining chunk */
+    assert(current >= pending_start);
+    copy_to_buf_len(buf, string + pending_start, current - pending_start);
+
+    copy_to_buf(buf, "\"");
+    return true;
+}
+
+#define write_buf_string_literal(buf, string) do { \
+    if (!check_write_string_literal(buf, string)) \
+        return false; \
+} while (0)
+
 static bool
 write_vmods(struct xkb_keymap *keymap, struct buf *buf)
 {
@@ -97,30 +199,28 @@ write_vmods(struct xkb_keymap *keymap, struct buf *buf)
     xkb_mod_index_t vmod = 0;
     bool has_some = false;
 
-    xkb_mods_enumerate(vmod, mod, &keymap->mods) {
-        if (mod->type != MOD_VIRT)
-            continue;
-
+    xkb_vmods_enumerate(vmod, mod, &keymap->mods) {
         if (!has_some) {
-            write_buf(buf, "\tvirtual_modifiers ");
+            copy_to_buf(buf, "\tvirtual_modifiers ");
             has_some = true;
         } else {
-            write_buf(buf, ",");
+            copy_to_buf(buf, ",");
         }
         write_buf(buf, "%s", xkb_atom_text(keymap->ctx, mod->name));
-        if (keymap->mods.explicit_vmods & (UINT32_C(1) << vmod)) {
-            /* Can only pretty-print real modifiers in this context */
-            if (mod->mapping & ~MOD_REAL_MASK_ALL) {
-                write_buf(buf, "=0x%"PRIx32, mod->mapping);
-            } else {
-                write_buf(buf, "=%s",
-                          ModMaskText(keymap->ctx, &keymap->mods, mod->mapping));
-            }
+        if ((keymap->mods.explicit_vmods & (UINT32_C(1) << vmod)) &&
+            mod->mapping != 0) {
+            /*
+             * Explicit non-default mapping
+             * NOTE: we can only pretty-print *real* modifiers in this context.
+             */
+            write_buf(buf, "=%s",
+                      ModMaskText(keymap->ctx, MOD_REAL, &keymap->mods,
+                                  mod->mapping));
         }
     }
 
     if (has_some)
-        write_buf(buf, ";\n\n");
+        copy_to_buf(buf, ";\n\n");
 
     return true;
 }
@@ -136,35 +236,37 @@ write_keycodes(struct xkb_keymap *keymap, struct buf *buf)
         write_buf(buf, "xkb_keycodes \"%s\" {\n",
                   keymap->keycodes_section_name);
     else
-        write_buf(buf, "xkb_keycodes {\n");
+        copy_to_buf(buf, "xkb_keycodes {\n");
 
     /* xkbcomp and X11 really want to see keymaps with a minimum of 8, and
      * a maximum of at least 255, else XWayland really starts hating life.
      * If this is a problem and people really need strictly bounded keymaps,
      * we should probably control this with a flag. */
-    write_buf(buf, "\tminimum = %u;\n", MIN(keymap->min_key_code, 8));
-    write_buf(buf, "\tmaximum = %u;\n", MAX(keymap->max_key_code, 255));
+    write_buf(buf, "\tminimum = %"PRIu32";\n", MIN(keymap->min_key_code, 8));
+    write_buf(buf, "\tmaximum = %"PRIu32";\n", MAX(keymap->max_key_code, 255));
 
     xkb_keys_foreach(key, keymap) {
         if (key->name == XKB_ATOM_NONE)
             continue;
 
-        write_buf(buf, "\t%-20s = %u;\n",
+        write_buf(buf, "\t%-20s = %"PRIu32";\n",
                   KeyNameText(keymap->ctx, key->name), key->keycode);
     }
 
     xkb_leds_enumerate(idx, led, keymap)
-        if (led->name != XKB_ATOM_NONE)
-            write_buf(buf, "\tindicator %u = \"%s\";\n",
-                      idx + 1, xkb_atom_text(keymap->ctx, led->name));
+        if (led->name != XKB_ATOM_NONE) {
+            write_buf(buf, "\tindicator %u = ", idx + 1);
+            write_buf_string_literal(buf, xkb_atom_text(keymap->ctx, led->name));
+            copy_to_buf(buf, ";\n");
+        }
 
 
-    for (unsigned i = 0; i < keymap->num_key_aliases; i++)
+    for (darray_size_t i = 0; i < keymap->num_key_aliases; i++)
         write_buf(buf, "\talias %-14s = %s;\n",
                   KeyNameText(keymap->ctx, keymap->key_aliases[i].alias),
                   KeyNameText(keymap->ctx, keymap->key_aliases[i].real));
 
-    write_buf(buf, "};\n\n");
+    copy_to_buf(buf, "};\n\n");
     return true;
 }
 
@@ -175,20 +277,23 @@ write_types(struct xkb_keymap *keymap, struct buf *buf)
         write_buf(buf, "xkb_types \"%s\" {\n",
                   keymap->types_section_name);
     else
-        write_buf(buf, "xkb_types {\n");
+        copy_to_buf(buf, "xkb_types {\n");
 
-    write_vmods(keymap, buf);
+    if (!write_vmods(keymap, buf))
+        return false;
 
-    for (unsigned i = 0; i < keymap->num_types; i++) {
+    for (darray_size_t i = 0; i < keymap->num_types; i++) {
         const struct xkb_key_type *type = &keymap->types[i];
 
-        write_buf(buf, "\ttype \"%s\" {\n",
-                  xkb_atom_text(keymap->ctx, type->name));
+        copy_to_buf(buf, "\ttype ");
+        write_buf_string_literal(buf, xkb_atom_text(keymap->ctx, type->name));
+        copy_to_buf(buf, " {\n");
 
         write_buf(buf, "\t\tmodifiers= %s;\n",
-                  ModMaskText(keymap->ctx, &keymap->mods, type->mods.mods));
+                  ModMaskText(keymap->ctx, MOD_BOTH, &keymap->mods,
+                              type->mods.mods));
 
-        for (unsigned j = 0; j < type->num_entries; j++) {
+        for (darray_size_t j = 0; j < type->num_entries; j++) {
             const char *str;
             const struct xkb_key_type_entry *entry = &type->entries[j];
 
@@ -199,25 +304,30 @@ write_types(struct xkb_keymap *keymap, struct buf *buf)
             if (entry->level == 0 && entry->preserve.mods == 0)
                 continue;
 
-            str = ModMaskText(keymap->ctx, &keymap->mods, entry->mods.mods);
-            write_buf(buf, "\t\tmap[%s]= %u;\n",
+            str = ModMaskText(keymap->ctx, MOD_BOTH, &keymap->mods,
+                              entry->mods.mods);
+            write_buf(buf, "\t\tmap[%s]= %"PRIu32";\n",
                       str, entry->level + 1);
 
             if (entry->preserve.mods)
                 write_buf(buf, "\t\tpreserve[%s]= %s;\n",
-                          str, ModMaskText(keymap->ctx, &keymap->mods,
+                          str, ModMaskText(keymap->ctx, MOD_BOTH, &keymap->mods,
                                            entry->preserve.mods));
         }
 
         for (xkb_level_index_t n = 0; n < type->num_level_names; n++)
-            if (type->level_names[n])
-                write_buf(buf, "\t\tlevel_name[%u]= \"%s\";\n", n + 1,
-                          xkb_atom_text(keymap->ctx, type->level_names[n]));
+            if (type->level_names[n]) {
+                write_buf(buf, "\t\tlevel_name[%"PRIu32"]= ", n + 1);
+                write_buf_string_literal(
+                    buf, xkb_atom_text(keymap->ctx, type->level_names[n]));
+                copy_to_buf(buf, ";\n");
+            }
 
-        write_buf(buf, "\t};\n");
+
+        copy_to_buf(buf, "\t};\n");
     }
 
-    write_buf(buf, "};\n\n");
+    copy_to_buf(buf, "};\n\n");
     return true;
 }
 
@@ -225,13 +335,15 @@ static bool
 write_led_map(struct xkb_keymap *keymap, struct buf *buf,
               const struct xkb_led *led)
 {
-    write_buf(buf, "\tindicator \"%s\" {\n",
-              xkb_atom_text(keymap->ctx, led->name));
+    copy_to_buf(buf, "\tindicator ");
+    write_buf_string_literal(buf, xkb_atom_text(keymap->ctx, led->name));
+    copy_to_buf(buf, " {\n");
 
     if (led->which_groups) {
         if (led->which_groups != XKB_STATE_LAYOUT_EFFECTIVE) {
             write_buf(buf, "\t\twhichGroupState= %s;\n",
-                      LedStateMaskText(keymap->ctx, led->which_groups));
+                      LedStateMaskText(keymap->ctx, groupComponentMaskNames,
+                                       led->which_groups));
         }
         write_buf(buf, "\t\tgroups= 0x%02x;\n",
                   led->groups);
@@ -240,10 +352,12 @@ write_led_map(struct xkb_keymap *keymap, struct buf *buf,
     if (led->which_mods) {
         if (led->which_mods != XKB_STATE_MODS_EFFECTIVE) {
             write_buf(buf, "\t\twhichModState= %s;\n",
-                      LedStateMaskText(keymap->ctx, led->which_mods));
+                      LedStateMaskText(keymap->ctx, modComponentMaskNames,
+                                       led->which_mods));
         }
         write_buf(buf, "\t\tmodifiers= %s;\n",
-                  ModMaskText(keymap->ctx, &keymap->mods, led->mods.mods));
+                  ModMaskText(keymap->ctx, MOD_BOTH, &keymap->mods,
+                              led->mods.mods));
     }
 
     if (led->ctrls) {
@@ -251,7 +365,7 @@ write_led_map(struct xkb_keymap *keymap, struct buf *buf,
                   ControlMaskText(keymap->ctx, led->ctrls));
     }
 
-    write_buf(buf, "\t};\n");
+    copy_to_buf(buf, "\t};\n");
     return true;
 }
 
@@ -297,7 +411,7 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
         if (action->mods.flags & ACTION_MODS_LOOKUP_MODMAP)
             args = "modMapMods";
         else
-            args = ModMaskText(keymap->ctx, &keymap->mods,
+            args = ModMaskText(keymap->ctx, MOD_BOTH, &keymap->mods,
                                action->mods.mods.mods);
         write_buf(buf, "%s%s(modifiers=%s%s%s%s)%s", prefix, type, args,
                   (action->type != ACTION_TYPE_MOD_LOCK && (action->mods.flags & ACTION_LOCK_CLEAR)) ? ",clearLocks" : "",
@@ -309,7 +423,7 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
     case ACTION_TYPE_GROUP_SET:
     case ACTION_TYPE_GROUP_LATCH:
     case ACTION_TYPE_GROUP_LOCK:
-        write_buf(buf, "%s%s(group=%s%d%s%s)%s", prefix, type,
+        write_buf(buf, "%s%s(group=%s%"PRId32"%s%s)%s", prefix, type,
                   (!(action->group.flags & ACTION_ABSOLUTE_SWITCH) && action->group.group > 0) ? "+" : "",
                   (action->group.flags & ACTION_ABSOLUTE_SWITCH) ? action->group.group + 1 : action->group.group,
                   (action->type != ACTION_TYPE_GROUP_LOCK && (action->group.flags & ACTION_LOCK_CLEAR)) ? ",clearLocks" : "",
@@ -337,11 +451,11 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
     case ACTION_TYPE_PTR_BUTTON:
         write_buf(buf, "%s%s(button=", prefix, type);
         if (action->btn.button > 0 && action->btn.button <= 5)
-            write_buf(buf, "%d", action->btn.button);
+            write_buf(buf, "%u", action->btn.button);
         else
-            write_buf(buf, "default");
+            copy_to_buf(buf, "default");
         if (action->btn.count)
-            write_buf(buf, ",count=%d", action->btn.count);
+            write_buf(buf, ",count=%u", action->btn.count);
         if (args)
             write_buf(buf, "%s", args);
         write_buf(buf, ")%s", suffix);
@@ -375,7 +489,21 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
         write_buf(buf, "%sNoAction()%s", prefix, suffix);
         break;
 
+    case ACTION_TYPE_VOID:
+        /*
+         * VoidAction() is a libxkbcommon extension.
+         * Use LockControls as a backward-compatible fallback.
+         * We cannot serialize it to `NoAction()`, as it would be dropped in
+         * e.g. the context of multiple actions.
+         * We better not use `Private` either, because it could still be
+         * interpreted by X11.
+         */
+        write_buf(buf, "%sLockControls(controls=none,affect=neither)%s", prefix, suffix);
+        break;
+
     default:
+        /* Unsupported legacy actions should have degraded to NoAction */
+        assert(action->type != ACTION_TYPE_UNSUPPORTED_LEGACY);
         write_buf(buf,
                   "%s%s(type=0x%02x,data[0]=0x%02x,data[1]=0x%02x,data[2]=0x%02x,data[3]=0x%02x,data[4]=0x%02x,data[5]=0x%02x,data[6]=0x%02x)%s",
                   prefix, type, action->type, action->priv.data[0],
@@ -397,14 +525,14 @@ write_actions(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
 
     for (xkb_level_index_t level = 0; level < XkbKeyNumLevels(key, group);
          level++) {
-        const union xkb_action *actions;
-        unsigned int count;
 
         if (level != 0)
-            write_buf(buf, ", ");
+            copy_to_buf(buf, ", ");
 
-        count = xkb_keymap_key_get_actions_by_level(keymap, key->keycode,
-                                                    group, level, &actions);
+        const union xkb_action *actions = NULL;
+        xkb_action_count_t count = xkb_keymap_key_get_actions_by_level(
+            keymap, key, group, level, &actions
+        );
         buf2->size = 0;
         if (count == 0) {
             if (!write_action(keymap, buf2, &noAction, NULL, NULL))
@@ -417,10 +545,10 @@ write_actions(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
             write_buf(buf, "%*s", ACTION_PADDING, buf2->buf);
         }
         else {
-            write_buf(buf2, "{ ");
-            for (unsigned int k = 0; k < count; k++) {
+            copy_to_buf(buf2, "{ ");
+            for (xkb_action_count_t k = 0; k < count; k++) {
                 if (k != 0)
-                    write_buf(buf2, ", ");
+                    copy_to_buf(buf2, ", ");
                 size_t old_size = buf2->size;
                 if (!write_action(keymap, buf2, &(actions[k]), NULL, NULL))
                     return false;
@@ -434,7 +562,7 @@ write_actions(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
                 if (!write_action(keymap, buf2, &(actions[k]), NULL, NULL))
                     return false;
             }
-            write_buf(buf2, " }");
+            copy_to_buf(buf2, " }");
             write_buf(buf, "%*s", ACTION_PADDING, buf2->buf);
         }
     }
@@ -451,42 +579,80 @@ write_compat(struct xkb_keymap *keymap, struct buf *buf)
         write_buf(buf, "xkb_compatibility \"%s\" {\n",
                   keymap->compat_section_name);
     else
-        write_buf(buf, "xkb_compatibility {\n");
+        copy_to_buf(buf, "xkb_compatibility {\n");
 
-    write_vmods(keymap, buf);
+    if (!write_vmods(keymap, buf))
+        return false;
 
-    write_buf(buf, "\tinterpret.useModMapMods= AnyLevel;\n");
-    write_buf(buf, "\tinterpret.repeat= False;\n");
+    copy_to_buf(buf, "\tinterpret.useModMapMods= AnyLevel;\n");
+    copy_to_buf(buf, "\tinterpret.repeat= False;\n");
 
-    for (unsigned i = 0; i < keymap->num_sym_interprets; i++) {
-        const struct xkb_sym_interpret *si = &keymap->sym_interprets[i];
+    /* xkbcomp requires at least one interpret entry. */
+    const darray_size_t num_sym_interprets = keymap->num_sym_interprets
+                                           ? keymap->num_sym_interprets
+                                           : 1;
+    const struct xkb_sym_interpret* const sym_interprets
+        = keymap->num_sym_interprets
+        ? keymap->sym_interprets
+        : &default_interpret;
 
-        write_buf(buf, "\tinterpret %s+%s(%s) {\n",
+    for (darray_size_t i = 0; i < num_sym_interprets; i++) {
+        const struct xkb_sym_interpret *si = &sym_interprets[i];
+
+        write_buf(buf, "\tinterpret %s+%s(%s) {",
                   si->sym ? KeysymText(keymap->ctx, si->sym) : "Any",
                   SIMatchText(si->match),
-                  ModMaskText(keymap->ctx, &keymap->mods, si->mods));
+                  ModMaskText(keymap->ctx, MOD_BOTH, &keymap->mods, si->mods));
 
-        if (si->virtual_mod != XKB_MOD_INVALID)
-            write_buf(buf, "\t\tvirtualModifier= %s;\n",
+        bool has_explicit_properties = false;
+
+        if (si->virtual_mod != XKB_MOD_INVALID) {
+            write_buf(buf, "\n\t\tvirtualModifier= %s;",
                       ModIndexText(keymap->ctx, &keymap->mods,
                                    si->virtual_mod));
+            has_explicit_properties = true;
+        }
 
-        if (si->level_one_only)
-            write_buf(buf, "\t\tuseModMapMods=level1;\n");
+        if (si->level_one_only) {
+            copy_to_buf(buf, "\n\t\tuseModMapMods=level1;");
+            has_explicit_properties = true;
+        }
 
-        if (si->repeat)
-            write_buf(buf, "\t\trepeat= True;\n");
+        if (si->repeat) {
+            copy_to_buf(buf, "\n\t\trepeat= True;");
+            has_explicit_properties = true;
+        }
 
-        write_action(keymap, buf, &si->action, "\t\taction= ", ";\n");
-        write_buf(buf, "\t};\n");
+        if (si->num_actions > 1) {
+            copy_to_buf(buf, "\n\t\taction= {");
+            const char suffix[] = ", ";
+            for (xkb_action_count_t k = 0; k < si->num_actions; k++) {
+                if (!write_action(keymap, buf, &si->a.actions[k], "", suffix))
+                    return false;
+            }
+            buf->size -= sizeof(suffix) - 1; /* trailing comma */
+            copy_to_buf(buf, "};");
+            has_explicit_properties = true;
+        } else if (si->num_actions == 1) {
+            if (!write_action(keymap, buf, &si->a.action,
+                              "\n\t\taction= ", ";"))
+                return false;
+            has_explicit_properties = true;
+        }
+        write_buf(buf, (has_explicit_properties
+                        ? "\n\t};\n"
+                        /* Empty interpret is a syntax error in xkbcomp, so
+                         * use a dummy entry */
+                        : "\n\t\taction= NoAction();\n\t};\n"));
     }
 
     xkb_leds_foreach(led, keymap)
         if (led->which_groups || led->groups || led->which_mods ||
             led->mods.mods || led->ctrls)
-            write_led_map(keymap, buf, led);
+            if (!write_led_map(keymap, buf, led))
+                return false;
 
-    write_buf(buf, "};\n\n");
+    copy_to_buf(buf, "};\n\n");
 
     return true;
 }
@@ -503,7 +669,7 @@ write_keysyms(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
         int num_syms;
 
         if (level != 0)
-            write_buf(buf, ", ");
+            copy_to_buf(buf, ", ");
 
         num_syms = xkb_keymap_key_get_syms_by_level(keymap, key->keycode,
                                                     group, level, &syms);
@@ -515,14 +681,14 @@ write_keysyms(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
         }
         else {
             buf2->size = 0;
-            write_buf(buf2, "{ ");
+            copy_to_buf(buf2, "{ ");
             for (int s = 0; s < num_syms; s++) {
                 if (s != 0)
-                    write_buf(buf2, ", ");
+                    copy_to_buf(buf2, ", ");
                 write_buf(buf2, "%*s", (show_actions ? padding : 0),
                           KeysymText(keymap->ctx, syms[s]));
             }
-            write_buf(buf2, " }");
+            copy_to_buf(buf2, " }");
             write_buf(buf, "%*s", padding, buf2->buf);
         }
     }
@@ -534,41 +700,40 @@ static bool
 write_key(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
           const struct xkb_key *key)
 {
-    xkb_layout_index_t group;
     bool simple = true;
-    bool explicit_types = false;
-    bool multi_type = false;
-    bool show_actions;
 
     write_buf(buf, "\tkey %-20s {", KeyNameText(keymap->ctx, key->name));
 
-    for (group = 0; group < key->num_groups; group++) {
-        if (key->groups[group].explicit_type)
-            explicit_types = true;
-
-        if (group != 0 && key->groups[group].type != key->groups[0].type)
-            multi_type = true;
-    }
-
-    if (explicit_types) {
-        const struct xkb_key_type *type;
+    if (key->explicit & EXPLICIT_TYPES) {
         simple = false;
 
+        bool multi_type = false;
+        for (xkb_layout_index_t group = 1; group < key->num_groups; group++) {
+            if (key->groups[group].type != key->groups[0].type) {
+                multi_type = true;
+                break;
+            }
+        }
+
         if (multi_type) {
-            for (group = 0; group < key->num_groups; group++) {
+            for (xkb_layout_index_t group = 0; group < key->num_groups; group++) {
                 if (!key->groups[group].explicit_type)
                     continue;
 
-                type = key->groups[group].type;
-                write_buf(buf, "\n\t\ttype[Group%u]= \"%s\",",
-                            group + 1,
-                            xkb_atom_text(keymap->ctx, type->name));
+                const struct xkb_key_type * const type = key->groups[group].type;
+                /* TODO: This will require using integer indexes when > 4 */
+                write_buf(buf, "\n\t\ttype[Group%"PRIu32"]= ", group + 1);
+                write_buf_string_literal(
+                  buf, xkb_atom_text(keymap->ctx, type->name));
+                copy_to_buf(buf, ",");
             }
         }
         else {
-            type = key->groups[0].type;
-            write_buf(buf, "\n\t\ttype= \"%s\",",
-                        xkb_atom_text(keymap->ctx, type->name));
+            const struct xkb_key_type * const type = key->groups[0].type;
+            write_buf(buf, "\n\t\ttype= ");
+            write_buf_string_literal(
+                buf, xkb_atom_text(keymap->ctx, type->name));
+            copy_to_buf(buf, ",");
         }
     }
 
@@ -581,31 +746,32 @@ write_key(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
      * One side effect is that no interpretation will be run on this key anymore,
      * so we may have to set some extra fields explicitly: repeat, virtualMods.
      */
-    show_actions = (key->explicit & EXPLICIT_INTERP);
+    const bool show_actions = (key->explicit & EXPLICIT_INTERP);
 
     /* If we show actions, interprets are not going to be used to set this
      * field, so make it explicit. */
     if ((key->explicit & EXPLICIT_REPEAT) || show_actions) {
         if (key->repeats)
-            write_buf(buf, "\n\t\trepeat= Yes,");
+            copy_to_buf(buf, "\n\t\trepeat= Yes,");
         else
-            write_buf(buf, "\n\t\trepeat= No,");
-        simple = false;
+            copy_to_buf(buf, "\n\t\trepeat= No,");
     }
 
     /* If we show actions, interprets are not going to be used to set this
      * field, so make it explicit. */
-    if (key->vmodmap && ((key->explicit & EXPLICIT_VMODMAP) || show_actions))
+    if ((key->explicit & EXPLICIT_VMODMAP) || (show_actions && key->vmodmap))
         write_buf(buf, "\n\t\tvirtualMods= %s,",
-                  ModMaskText(keymap->ctx, &keymap->mods, key->vmodmap));
+                  ModMaskText(keymap->ctx, MOD_BOTH, &keymap->mods,
+                              key->vmodmap));
 
     switch (key->out_of_range_group_action) {
     case RANGE_SATURATE:
-        write_buf(buf, "\n\t\tgroupsClamp,");
+        copy_to_buf(buf, "\n\t\tgroupsClamp,");
         break;
 
     case RANGE_REDIRECT:
-        write_buf(buf, "\n\t\tgroupsRedirect= Group%u,",
+        /* TODO: This will require using integer indexes when > 4 */
+        write_buf(buf, "\n\t\tgroupsRedirect= Group%"PRIu32",",
                     key->out_of_range_group_number + 1);
         break;
 
@@ -617,27 +783,40 @@ write_key(struct xkb_keymap *keymap, struct buf *buf, struct buf *buf2,
         simple = false;
 
     if (simple) {
-        write_buf(buf, "\t[ ");
-        if (!write_keysyms(keymap, buf, buf2, key, 0, false))
-            return false;
-        write_buf(buf, " ] };\n");
+        const bool only_symbols = key->explicit == EXPLICIT_SYMBOLS;
+        if (key->num_groups == 0) {
+            /* Remove trailing comma */
+            if (buf->buf[buf->size - 1] == ',')
+                buf->size--;
+        } else {
+	        if (!only_symbols)
+	            copy_to_buf(buf, "\n\t");
+            copy_to_buf(buf, "\t[ ");
+            if (!write_keysyms(keymap, buf, buf2, key, 0, false))
+                return false;
+            copy_to_buf(buf, " ]");
+        }
+        write_buf(buf, "%s", (only_symbols ? " };\n" : "\n\t};\n"));
     }
     else {
-        for (group = 0; group < key->num_groups; group++) {
+        assert(key->num_groups > 0);
+        for (xkb_layout_index_t group = 0; group < key->num_groups; group++) {
             if (group != 0)
-                write_buf(buf, ",");
-            write_buf(buf, "\n\t\tsymbols[Group%u]= [ ", group + 1);
+                copy_to_buf(buf, ",");
+            /* TODO: This will require using integer indexes when > 4 */
+            write_buf(buf, "\n\t\tsymbols[Group%"PRIu32"]= [ ", group + 1);
             if (!write_keysyms(keymap, buf, buf2, key, group, show_actions))
                 return false;
-            write_buf(buf, " ]");
+            copy_to_buf(buf, " ]");
             if (show_actions) {
-                write_buf(buf, ",\n\t\tactions[Group%u]= [ ", group + 1);
+                /* TODO: This will require using integer indexes when > 4 */
+                write_buf(buf, ",\n\t\tactions[Group%"PRIu32"]= [ ", group + 1);
                 if (!write_actions(keymap, buf, buf2, key, group))
                     return false;
-                write_buf(buf, " ]");
+                copy_to_buf(buf, " ]");
             }
         }
-        write_buf(buf, "\n\t};\n");
+        copy_to_buf(buf, "\n\t};\n");
     }
 
     return true;
@@ -655,19 +834,23 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
         write_buf(buf, "xkb_symbols \"%s\" {\n",
                   keymap->symbols_section_name);
     else
-        write_buf(buf, "xkb_symbols {\n");
+        copy_to_buf(buf, "xkb_symbols {\n");
 
     for (group = 0; group < keymap->num_group_names; group++)
-        if (keymap->group_names[group])
-            write_buf(buf,
-                      "\tname[Group%u]=\"%s\";\n", group + 1,
-                      xkb_atom_text(keymap->ctx, keymap->group_names[group]));
+        if (keymap->group_names[group]) {
+            /* TODO: This will require using integer indexes when > 4 */
+            write_buf(buf, "\tname[Group%"PRIu32"]=", group + 1);
+            write_buf_string_literal(
+                buf, xkb_atom_text(keymap->ctx, keymap->group_names[group]));
+            copy_to_buf(buf, ";\n");
+        }
     if (group > 0)
-        write_buf(buf, "\n");
+        copy_to_buf(buf, "\n");
 
     struct buf buf2 = { NULL, 0, 0 };
     xkb_keys_foreach(key, keymap) {
-        if (key->num_groups > 0) {
+        /* Skip keys with no explicit values */
+        if (key->explicit) {
             if (!write_key(keymap, buf, &buf2, key)) {
                 free(buf2.buf);
                 return false;
@@ -676,7 +859,7 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
     }
     free(buf2.buf);
 
-    xkb_mods_enumerate(i, mod, &keymap->mods) {
+    xkb_rmods_enumerate(i, mod, &keymap->mods) {
         bool had_any = false;
         xkb_keys_foreach(key, keymap) {
             if (key->modmap & (UINT32_C(1) << i)) {
@@ -690,10 +873,10 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
             }
         }
         if (had_any)
-            write_buf(buf, " };\n");
+            copy_to_buf(buf, " };\n");
     }
 
-    write_buf(buf, "};\n\n");
+    copy_to_buf(buf, "};\n\n");
     return true;
 }
 
