@@ -1,24 +1,6 @@
 /*
  * Copyright © 2012 Ran Benita <ran234@gmail.com>
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  */
 
 #include "config.h"
@@ -971,7 +953,7 @@ test_explicit_actions(struct xkb_context *ctx)
 
     struct xkb_keymap *keymaps[] = { original, roundtrip };
 
-    for (unsigned k = 0; k < ARRAY_SIZE(keymaps); k++) {
+    for (unsigned int k = 0; k < ARRAY_SIZE(keymaps); k++) {
         assert(keymaps[k]);
 
         /*
@@ -992,7 +974,7 @@ test_explicit_actions(struct xkb_context *ctx)
             /* No explicit actions, check defaults */
             { .name = "AD06", .repeats = true , .vmodmap = 0 },
         };
-        for (unsigned i = 0; i < ARRAY_SIZE(keys); i++) {
+        for (unsigned int i = 0; i < ARRAY_SIZE(keys); i++) {
             xkb_keycode_t kc = xkb_keymap_key_by_name(keymaps[k], keys[i].name);
             assert(kc != XKB_KEYCODE_INVALID);
             assert(keys[i].repeats == xkb_keymap_key_repeats(keymaps[k], kc));
@@ -1102,20 +1084,13 @@ test_simultaneous_modifier_clear(struct xkb_context *context)
     xkb_keymap_unref(keymap);
 }
 
-int
-main(void)
+static void
+test_keymaps(struct xkb_context *ctx, const char* rules)
 {
-    test_init();
-
-    struct xkb_context *ctx = test_get_context(CONTEXT_NO_FLAG);
     struct xkb_keymap *keymap;
 
     assert(ctx);
 
-    test_simultaneous_modifier_clear(ctx);
-    test_group_latch(ctx);
-    test_mod_latch(ctx);
-    test_explicit_actions(ctx);
 
     keymap = test_compile_rules(ctx, "evdev", "evdev",
                                 "us,il,ru,de", ",,phonetic,neo",
@@ -1394,7 +1369,7 @@ main(void)
                         KEY_V,           BOTH,  XKB_KEY_p,               FINISH));
 
     xkb_keymap_unref(keymap);
-    keymap = test_compile_rules(ctx, "evdev", "", "de", "neo", "");
+    keymap = test_compile_rules(ctx, rules, "", "de", "neo", "");
     assert(keymap);
     assert(test_key_seq(keymap,
                         /* Level 5. */
@@ -1446,7 +1421,7 @@ main(void)
 
 
     xkb_keymap_unref(keymap);
-    keymap = test_compile_rules(ctx, "evdev", "", "us,il,ru", "",
+    keymap = test_compile_rules(ctx, rules, "", "us,il,ru", "",
                                 "grp:alt_shift_toggle_bidir,grp:menu_toggle");
     assert(keymap);
 
@@ -1485,7 +1460,7 @@ main(void)
                         KEY_H,         BOTH, XKB_KEY_h,              FINISH));
 
     xkb_keymap_unref(keymap);
-    keymap = test_compile_rules(ctx, "evdev", "", "us,il,ru", "",
+    keymap = test_compile_rules(ctx, rules, "", "us,il,ru", "",
                                 "grp:switch,grp:lswitch,grp:menu_toggle");
     assert(keymap);
 
@@ -1528,7 +1503,7 @@ main(void)
                         KEY_H,         BOTH, XKB_KEY_h,                 FINISH));
 
     xkb_keymap_unref(keymap);
-    keymap = test_compile_rules(ctx, "evdev", "", "us", "euro", "");
+    keymap = test_compile_rules(ctx, rules, "", "us", "euro", "");
     assert(keymap);
 
     assert(test_key_seq(keymap,
@@ -1548,7 +1523,7 @@ main(void)
                         KEY_Z,         BOTH, XKB_KEY_y,                 FINISH));
 
     xkb_keymap_unref(keymap);
-    keymap = test_compile_rules(ctx, "evdev", "applealu_ansi", "us", "",
+    keymap = test_compile_rules(ctx, rules, "applealu_ansi", "us", "",
                                 "terminate:ctrl_alt_bksp");
     assert(keymap);
 
@@ -1568,6 +1543,26 @@ main(void)
                         KEY_A,         BOTH, XKB_KEY_a,                 FINISH));
 
     xkb_keymap_unref(keymap);
+}
+
+int
+main(void)
+{
+    test_init();
+
+    struct xkb_context *ctx = test_get_context(CONTEXT_NO_FLAG);
+    assert(ctx);
+
+    /* Usual rules */
+    test_keymaps(ctx, "evdev");
+    /* Special rules to make no use of modmaps */
+    test_keymaps(ctx, "evdev-pure-virtual-mods");
+
+    test_simultaneous_modifier_clear(ctx);
+    test_group_latch(ctx);
+    test_mod_latch(ctx);
+    test_explicit_actions(ctx);
+
     xkb_context_unref(ctx);
     return EXIT_SUCCESS;
 }
