@@ -239,7 +239,7 @@ buffer_create(struct interactive_dpy *inter, uint32_t width, uint32_t height)
         stride = width * 2;
         break;
     default:
-        fprintf(stderr, "Unsupported SHM format %d\n", inter->shm_format);
+        fprintf(stderr, "Unsupported SHM format %"PRIu32"\n", inter->shm_format);
         exit(EXIT_FAILURE);
     }
 
@@ -615,7 +615,7 @@ seat_create(struct interactive_dpy *inter, struct wl_registry *registry,
         seat->compose_state = xkb_compose_state_new(seat->inter->compose_table,
                                                     XKB_COMPOSE_STATE_NO_FLAGS);
     }
-    ret = asprintf(&seat->name_str, "seat:%d",
+    ret = asprintf(&seat->name_str, "seat:%"PRIu32,
                    wl_proxy_get_id((struct wl_proxy *) seat->wl_seat));
     assert(ret >= 0);
     wl_list_insert(&inter->seats, &seat->link);
@@ -771,6 +771,8 @@ main(int argc, char *argv[])
         {0, 0, 0, 0},
     };
 
+    setlocale(LC_ALL, "");
+
     while (1) {
         int opt;
         int option_index = 0;
@@ -798,8 +800,6 @@ main(int argc, char *argv[])
         }
     }
 
-    setlocale(LC_ALL, "");
-
     memset(&inter, 0, sizeof(inter));
     wl_list_init(&inter.seats);
 
