@@ -6,6 +6,7 @@
 #include "config.h"
 
 #include <assert.h>
+#include <locale.h>
 #include <stdio.h>
 #include <getopt.h>
 
@@ -50,6 +51,8 @@ main(int argc, char **argv)
         {0, 0, 0, 0},
     };
 
+    setlocale(LC_ALL, "");
+
     while (1) {
         int c;
         int option_index = 0;
